/* -*- c++ -*-

This is Picprog, Microchip PIC programmer software for the serial port device.
Copyright  1997,2002,2003,2004 Jaakko Hyvtti

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2 as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software Foundation,
Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

The author may be contacted at:

Email: Jaakko.Hyvatti@iki.fi
URL:   http://www.iki.fi/hyvatti/
Phone: +358 40 5011222

Please send any suggestions, bug reports, success stories etc. to the
Email address above.

*/

#include <iostream>
#include <iomanip>

#include <sys/ioctl.h>
#include <fcntl.h>
#include <errno.h>
#include <time.h>
#include <sys/time.h>
#include <stdlib.h>
#include <unistd.h>
#include <sysexits.h>
#include <string.h>

#include "picport.h"

using namespace std;

const int picport::dtr_bit = TIOCM_DTR;
const int picport::rts_bit = TIOCM_RTS;

#ifndef O_NONBLOCK
#define O_NONBLOCK O_NDELAY
#endif

picport::picport (const char *tty)
  : addr (0), debug_on (0), W(0)
{
  if (0 > (fd = open (tty, O_RDWR|O_NOCTTY|O_NONBLOCK))) {
    int e = errno;
    cerr << "Unable to open tty " << tty << ":" << strerror (e) << endl;
    exit (EX_IOERR);
  }
  tcgetattr (fd, &saved);
  termstate = saved;
  termstate.c_iflag = IGNBRK | IGNPAR;
  termstate.c_oflag = 0;
  termstate.c_cflag = B9600 | CS8 | CREAD | CLOCAL;
  termstate.c_lflag = 0;
  tcsetattr (fd, TCSANOW, &termstate);

  // Initialize lines for programming

  // RTS/DTR low, sleep and then TxD high (program voltage)

  ioctl (fd, TIOCCBRK, 0);
  if (0 > ioctl (fd, TIOCMBIC, &picport::rts_bit)
      || 0 > ioctl (fd, TIOCMBIC, &picport::dtr_bit)) {
    int e = errno;
    tcsetattr (fd, TCSANOW, &saved);
    cerr << "Unable to clear RTS/DTR on tty " << tty << ":" << strerror (e) << endl;
    exit (EX_IOERR);
  }
  usleep (50);
  // Check the CTS.  If it is up, even when we just lowered DTR,
  // we probably are not talking to a JDM type programmer.
  int i;
  ioctl (fd, TIOCMGET, &i);
  if (i & TIOCM_CTS) {
    tcsetattr (fd, TCSANOW, &saved);
    cerr << tty << ":CTS is high, probably we are not connected to a programmer but a modem or terminal." << endl;
    exit (EX_IOERR);
  }

  if (0 > ioctl (fd, TIOCSBRK, 0)) {
    int e = errno;
    ioctl (fd, TIOCCBRK, 0);
    tcsetattr (fd, TCSANOW, &saved);
    cerr << "Unable to start break on tty " << tty << ":" << strerror (e) << endl;
    exit (EX_IOERR);
  }
  usleep (10);
  portname = new char [strlen (tty) + 1];
  strcpy (portname, tty);
}

picport::~picport ()
{
  ioctl (fd, TIOCCBRK, 0);
  usleep (1);
  tcsetattr (fd, TCSANOW, &saved);
  close (fd);
  delete [] portname;
}

void picport::reset ()
{
  ioctl (fd, TIOCMBIC, &picport::rts_bit);
  ioctl (fd, TIOCMBIC, &picport::dtr_bit);
  ioctl (fd, TIOCCBRK, 0);
  usleep (50);
  ioctl (fd, TIOCSBRK, 0);
  usleep (10);
  addr = 0;
}

// for debugging purposes: put data bit up
void picport::force ()
{
  ioctl (fd, TIOCMBIS, &picport::dtr_bit);
}

// Change these defines if you want different timing methods used.  I
// have noticed that many implementations have these delay functions
// waiting unacceptably long, thus making the programming take minutes
// instead of seconds.
#undef NANOSLEEP_WORKS
#undef USLEEP_WORKS

static void delay_1000 ()
{
#ifdef NANOSLEEP_WORKS
  static const struct timespec ns1000 = { 0, 1000 };
  nanosleep (&ns1000, 0);
#else
#ifdef USLEEP_WORKS
  usleep (1);
#else
  // Delay loop that should take more than a microsecond to execute.
  // Check the real time clock also and break out if at least one
  // microsecond has gone.

  struct timeval tv1, tv2;
  volatile int i;
  gettimeofday (&tv1, 0);
  for (i = 0; i < 10000; i++) {
    gettimeofday (&tv2, 0);
    tv2.tv_sec -= tv1.tv_sec;
    if (0 > (tv2.tv_usec -= tv1.tv_usec)) {
      tv2.tv_sec--;
      tv2.tv_usec += 1000000;
    }
    if (tv2.tv_sec || tv2.tv_usec > 1)
      break;
  }
#endif
#endif
}

static void delay_100 ()
{
#ifdef NANOSLEEP_WORKS
  static const struct timespec ns100 = { 0, 100 };
  nanosleep (&ns100, 0);
#else
#ifdef USLEEP_WORKS
  usleep (1);
#else
  delay_1000 ();
#endif
#endif
}

void picport::p_out (int b)
{
  if (b)
    ioctl (fd, TIOCMBIS, &picport::dtr_bit); // set data
  ioctl (fd, TIOCMBIS, &picport::rts_bit); // clock up
  delay_100 ();
  ioctl (fd, TIOCMBIC, &picport::rts_bit); // clock down
  delay_100 ();
  if (b)
    ioctl (fd, TIOCMBIC, &picport::dtr_bit); // set data down
}

int picport::p_in ()
{
  ioctl (fd, TIOCMBIS, &picport::rts_bit); // clock up
  delay_100 ();
  ioctl (fd, TIOCMBIS, &picport::dtr_bit); // set data up
  ioctl (fd, TIOCMBIC, &picport::rts_bit); // clock down
  delay_100 ();

  int i;

  ioctl (fd, TIOCMGET, &i);
  ioctl (fd, TIOCMBIC, &picport::dtr_bit); // set data down
  return 0 != (i & TIOCM_CTS);
}

int picport::command18 (enum commands18 comm, int data)
{
  int i, shift = comm;
  if (nop_prog == comm) {
    // A programming command must leave the last bit clock pulse up
    p_out (0);
    p_out (0);
    p_out (0);
    usleep (1000); // Allow power supply to charge fully
    // data line is already 0, as needed
    ioctl (fd, TIOCMBIS, &picport::rts_bit); // clock up
    usleep (1000); // P9 >1 ms programming time
    ioctl (fd, TIOCMBIC, &picport::rts_bit); // clock down
    usleep (5); // P10 >5 s high voltage discharge time
  } else {
    for (i = 0; i < 4; i++)
      p_out ((shift >> i) & 1);
    delay_1000 ();
  }

  shift = 0; // default return value

  switch (comm) {
  case nop_erase:
    // Erase cycle has delay between command and data
    usleep (10000); // P11 5 ms + P10 5 s erase time
    // FALLTHROUGH
    
  case instr:
  case nop_prog:
    if (0x0e00 == (data & 0xff00))
      W = data & 0x00ff;
    else if (0x6ef8 == data)
      addr = (addr & 0x00ffff) | (W << 16);
    else if (0x6ef7 == data)
      addr = (addr & 0xff00ff) | (W << 8);
    else if (0x6ef6 == data)
      addr = (addr & 0xffff00) | W;
    goto sw;

  case twrite_dec2:
    addr -= 2;
    goto sw;

  case twrite_inc2:
    addr += 2;
    // FALLTHROUGH

  case twrite:
  case twrite_prog:
  sw:
    for (i = 0; i < 16; i++)
      p_out ((data >> i) & 1);
    break;

  case tread_dec:
    --addr;
    goto sr;

  case tread_inc:
  case inc_tread:
    ++addr;
    // FALLTHROUGH

  case shift_out:
  case tread:
  sr:
    for (i = 0; i < 8; i++)
      p_in();
    for (i = 0; i < 8; i++)
      shift |= p_in () << i;
  }

  delay_1000 ();
  return shift;
}

void picport::setaddress (unsigned long a)
{
  if (0 != a && addr == a)
    return;
  command18 (instr, 0x0e00 | ((a & 0xff0000) >> 16));
  command18 (instr, 0x6ef8);
  command18 (instr, 0x0e00 | ((a & 0x00ff00) >> 8));
  command18 (instr, 0x6ef7);
  command18 (instr, 0x0e00 | (a & 0x0000ff));
  command18 (instr, 0x6ef6);
}

// -1 == error, no programmer present

int picport::command (enum commands comm, int data)
{
  int tmp1, tmp2;

  // first, send out the command, 6 bits

  int i, shift = comm;
  for (i = 0; i < 6; i++)
    p_out ((shift >> i) & 1);

  shift = 0; // default return value

  switch (comm) {
  case inc_addr:
    if (++addr >= 0x4000)
      addr = 0x2000;
    break;

  case data_from_prog:
  case data_from_data:
    delay_1000 ();
    tmp1 = p_in ();
    for (i = 0; i < 14; i++)
      shift |= p_in () << i;
    tmp2 = p_in ();

    // Start and stop bits must be 1.

    if (!tmp1 || !tmp2) {
      cerr << portname << ":PIC programmer missing or chip fault" << endl;
      return -1;
    }

    if (data_from_data == comm) {

      // Check that the leftover bits were valid, all 1's.
      // This detects if the programmer is not connected to the port.
      // Unfortunately later chips clear these bits, so we must
      // accept both all 1's and all 0's.

      if ((shift & 0x3f00) != 0x3f00
	  && (shift & 0x3f00) != 0x0000) {
	cerr << portname << ": read value "
	     << hex << setfill('0') << setw(4) << shift << dec
	     << ": PIC programmer or chip fault\n"
	  "Is code protection enabled?  "
	  "Use --erase option to disable code protection." << endl;
	return -1;
      }

      shift &= 0xff;
    }
    break;

  case load_conf:
    addr = 0x2000;
    // FALLTHROUGH

  case data_for_prog:
  case data_for_data:
    delay_1000 ();
    p_out (0);
    for (i = 0; i < 14; i++)
      p_out ((data >> i) & 1);
    p_out (0);
    break;

  default:
    ;
  }

  delay_1000 ();
  return shift;
}

